# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from matplotlib.colors import LogNorm
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from warnings import warn

from spectre_core.plotting.base import BaseTimeSeriesPanel
from spectre_core.plotting.panel_register import register_panel
from spectre_core.plotting.base import CutsPanel
from spectre_core.plotting.library.time_cuts.panel import Panel as TimeCutsPanel
from spectre_core.plotting.library.frequency_cuts.panel import Panel as FrequencyCutsPanel
from spectre_core.plotting.format import DEFAULT_FORMATS

SPECTROGRAM_PANEL_NAME = "spectrogram"

@register_panel(SPECTROGRAM_PANEL_NAME)
class Panel(BaseTimeSeriesPanel):
    def __init__(self, 
                 *args, 
                 log_norm: bool = False,
                 dBb: bool = False,
                 vmin: float | None = -1,
                 vmax: float | None = 2,
                 cmap: str = DEFAULT_FORMATS.spectrogram_cmap,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._log_norm = log_norm
        self._dBb = dBb
        self._vmin = vmin
        self._vmax = vmax
        self._cmap = cmap


    def draw(self):
        dynamic_spectra = self._spectrogram.dynamic_spectra_as_dBb if self._dBb else self._spectrogram.dynamic_spectra

        norm = LogNorm(vmin=np.nanmin(dynamic_spectra[dynamic_spectra > 0]), 
                       vmax=np.nanmax(dynamic_spectra)) if self._log_norm else None
        

        if self._log_norm and (self._vmin or self._vmax):
            warn("vmin/vmax will be ignored while using log_norm.")
            self._vmin = None 
            self._vmax = None
        
        # Plot the spectrogram with kwargs
        pcm = self.ax.pcolormesh(self.times, 
                            self._spectrogram.frequencies * 1e-6, 
                            dynamic_spectra,
                            vmin=self._vmin, 
                            vmax=self._vmax,
                            norm=norm, 
                            cmap=self._cmap)
        
        # Add colorbar if dBb is used
        if self._dBb:
            cbar = self.fig.colorbar(pcm, 
                                      ax=self.ax, 
                                      ticks=np.linspace(self._vmin, self._vmax, 6, dtype=int))
            cbar.set_label('dBb')


    def annotate_y_axis(self) -> None:
        self.ax.set_ylabel('Frequency [MHz]')
        return
    
    
    def overlay_cuts(self, cuts_panel: CutsPanel) -> None:
        if isinstance(cuts_panel, TimeCutsPanel):
            self._overlay_time_cuts(cuts_panel)
        elif isinstance(cuts_panel, FrequencyCutsPanel):
            self._overlay_frequency_cuts(cuts_panel)


    def _overlay_time_cuts(self, time_cuts_panel: TimeCutsPanel) -> None:
        for frequency, color in time_cuts_panel.bind_to_colors():
            self.ax.axhline(frequency*1e-6, # convert to MHz
                            color = color,
                            linewidth=DEFAULT_FORMATS.line_width
                            )
            
            
    def _overlay_frequency_cuts(self, frequency_cuts_panel: FrequencyCutsPanel) -> None:
        for time, color in frequency_cuts_panel.bind_to_colors():
            self.ax.axvline(time,
                            color = color,
                            linewidth=DEFAULT_FORMATS.line_width
                            )



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from spectre_core.receivers.base import SDRPlayReceiver
from spectre_core.receivers.receiver_register import register_receiver
from spectre_core.receivers.library.rsp1a.gr import fixed, sweep
from spectre_core.file_handlers.configs import CaptureConfig

@register_receiver("rsp1a")
class Receiver(SDRPlayReceiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def _set_capture_methods(self) -> None:
        self._capture_methods = {
            "fixed": self.__fixed,
            "sweep": self.__sweep
        }
        return
    

    def _set_validators(self) -> None:
        self._validators = {
            "fixed": self.__fixed_validator,
            "sweep": self.__sweep_validator
        }
        return
    

    def _set_type_templates(self) -> None:
        self._type_templates = {
            "fixed": self._get_default_type_template("fixed"),
            "sweep": self._get_default_type_template("sweep")
        }


    def _set_specifications(self) -> None:
        self._specifications = {
            "center_freq_lower_bound": 1e3, # [Hz]
            "center_freq_upper_bound": 2e9, # [Hz]
            "samp_rate_lower_bound": 200e3, # [Hz]
            "samp_rate_upper_bound": 10e6, # [Hz]
            "bandwidth_lower_bound": 200e3, # [Hz]
            "bandwidth_upper_bound": 8e6, # [Hz]
            "IF_gain_upper_bound": -20, # [dB]
            "RF_gain_upper_bound": 0, # [dB]
            "api_latency": 50 * 1e-3 # [s]
        }


    def __fixed(self, capture_config: CaptureConfig) -> None:
        fixed.main(capture_config)
    
    
    def __sweep(self, capture_config: CaptureConfig) -> None:
        sweep.main(capture_config)
    

    def __fixed_validator(self, capture_config: CaptureConfig) -> None:
        self._default_fixed_validator(capture_config)
        self._sdrplay_validator(capture_config)
    

    def __sweep_validator(self, capture_config: CaptureConfig) -> None:
        self._default_sweep_validator(capture_config)
        self._sdrplay_validator(capture_config)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from spectre_core.dynamic_imports import import_target_modules

import_target_modules(__file__, __name__, "chunk")




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

from math import floor
import warnings

from scipy.signal import get_window


def closed_upper_bound_RF_gain(RF_gain: float, 
                               RF_gain_upper_bound: float) -> None:
    if not (RF_gain <= RF_gain_upper_bound):
        raise ValueError((f"RF gain must be less than or equal to {RF_gain_upper_bound} [dB]. "
                          f"Got {RF_gain} [dB]"))


def closed_upper_bound_IF_gain(IF_gain: float, 
                               IF_gain_upper_bound: float) -> None:
    if not (IF_gain <= IF_gain_upper_bound):
        raise ValueError((f"IF gain must be less than or equal to {IF_gain_upper_bound} [dB]. "
                          f"Got {IF_gain} [dB]"))


def closed_confine_center_freq(center_freq: float, 
                               center_freq_lower_bound: float, 
                               center_freq_upper_bound: float) -> None:
    if not (center_freq_lower_bound <= center_freq <= center_freq_upper_bound):
        raise ValueError((f"Center frequency must be between {center_freq_lower_bound*1e-3} [kHz] and {center_freq_upper_bound*1e-9} [GHz]. "
                          f"Got {center_freq*1e-6} [MHz]"))
    

def closed_confine_samp_rate(samp_rate: int, 
                             samp_rate_lower_bound: int, 
                             samp_rate_upper_bound: int) -> None:
    if not (samp_rate_lower_bound <= samp_rate <= samp_rate_upper_bound):
        raise ValueError((f"Sampling rate must be between {samp_rate_lower_bound*1e-6} [MHz] and {samp_rate_upper_bound*1e-6} [MHz]. "
                          f"Got {samp_rate*1e-6} [MHz]"))
    

def closed_confine_bandwidth(bandwidth: float, 
                             bandwidth_lower_bound: float, 
                             bandwidth_upper_bound: float) -> None:
    if not (bandwidth_lower_bound <= bandwidth <= bandwidth_upper_bound):
        raise ValueError((f"Bandwidth must be between {bandwidth_lower_bound*1e-3} [kHz] and {bandwidth_upper_bound*1e-6} [MHz]. "
                          f"Got {bandwidth*1e-6} [MHz]"))
    
    
def is_power_of_two(n):
    return (n != 0) and (n & (n-1) == 0)


def window(window_type: str, 
           window_kwargs: dict, 
           window_size: int,
           chunk_size: int,
           samp_rate: float) -> None:
    
    if not is_power_of_two(window_size):
        raise ValueError((f"Window size must be some power of two. "
                          f"Got {window_size} [samples]"))
    
    window_interval = window_size*(1/samp_rate)
    if window_interval > chunk_size:
        raise ValueError((f"The windowing interval must be strictly less than the chunk size. "
                          f"Computed the windowing interval to be {window_interval} [s], "
                          f"but the chunk size is {chunk_size} [s]"))
    
    try:
        window_params = (window_type, 
                         *window_kwargs.values())
        _ = get_window(window_params, window_size)
    except Exception as e:
        raise Exception((f"An error has occurred while validating the window. "
                         f"Got {str(e)}"))
    

def hop(hop: int):
    if hop < 0:
        raise ValueError((f"Window hop must be strictly positive. "
                          f"Got {hop} [samples]"))
    
    
def center_freq_strictly_positive(center_freq: float):
    if center_freq <= 0:
        raise ValueError((f"Center frequency must be strictly positive. "
                          f"Got {center_freq*1e-6} [MHz]"))
    

def bandwidth_strictly_positive(bandwidth: float) -> None:
    if bandwidth < 0:
        raise ValueError((f"Bandwidth must be non-negative. "
                          f"Got {bandwidth*1e-6} [MHz]"))
    

def nyquist_criterion(samp_rate: int, 
                      bandwidth: float) -> None:
    if samp_rate < bandwidth:
        raise ValueError((f"Sample rate must be greater than or equal to the bandwidth. "
                          f"Got sample rate {samp_rate} [Hz], and bandwidth {bandwidth} [Hz]"))
    

def samp_rate_strictly_positive(samp_rate: int) -> None:
    if samp_rate <= 0:
        raise ValueError((f"Sample rate must be strictly positive. "
                          f"Got {samp_rate} [Hz]"))
    

def chunk_size_strictly_positive(chunk_size: int) -> None:
    if chunk_size <= 0:
        raise ValueError((f"Chunk size must be strictly positive. "
                          f"Got {chunk_size} [s]"))
    

def time_resolution(time_resolution: float, 
                    chunk_size: int) -> None:
    if time_resolution < 0:
        raise ValueError((f"Time resolution must be non-negative. "
                          f"Got {time_resolution} [s]"))
    
    if time_resolution > chunk_size:
        raise ValueError(f"Time resolution must be less than or equal to chunk size. "
                         f"Got time resolution {time_resolution} [s], "
                         f"and chunk size {chunk_size} [s]")
    

def frequency_resolution(frequency_resolution: float,
                         bandwidth: float = None) -> None:
    if frequency_resolution < 0:
        raise ValueError((f"Frequency resolution must be non-negative. "
                          f"Got {frequency_resolution} [Hz]"))
    
    if bandwidth is not None and frequency_resolution >= bandwidth:
        raise ValueError((f"Frequency resolution must be less than the bandwidth. "
                          f"Got frequency resolution to be {frequency_resolution} [Hz], "
                          f"with bandwidth {bandwidth} [Hz]"))
    

def chunk_key(chunk_key: str, 
              expected_chunk_key: str) -> None:
    if chunk_key != expected_chunk_key:
        raise ValueError((f"Expected {expected_chunk_key} for the chunk key. "
                          f"Got {chunk_key}"))
    

def event_handler_key(event_handler_key: str, 
                      expected_event_handler_key: str) -> None:
    if event_handler_key != expected_event_handler_key:
        raise ValueError((f"Expected {expected_event_handler_key} for the event handler key. "
                          f"Got {event_handler_key}"))
    

def gain_is_negative(gain: float) -> None:
    if gain > 0:
        raise ValueError(f"Gain must be non-positive. "
                         f"Got {gain} [dB]")
    

def _compute_num_steps_per_sweep(min_freq: float, 
                                 max_freq: float,
                                 samp_rate: int,
                                 freq_step: float) -> int:
     return floor((max_freq - min_freq + samp_rate/2) / freq_step)


def num_steps_per_sweep(min_freq: float, 
                        max_freq: float, 
                        samp_rate: int, 
                        freq_step: float) -> None:
    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, 
                                                       max_freq, 
                                                       samp_rate, 
                                                       freq_step)
    if num_steps_per_sweep <= 1:
        raise ValueError((f"We need strictly greater than one step per sweep. "
                          f"Computed {num_steps_per_sweep} step per sweep"))
    

def sweep_interval(min_freq: float, 
                  max_freq: float, 
                  samp_rate: int, 
                  freq_step: float,
                  samples_per_step: int,
                  chunk_size: float) -> None: 
    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, 
                                                       max_freq, 
                                                       samp_rate, 
                                                       freq_step)
    num_samples_per_sweep = num_steps_per_sweep * samples_per_step
    sweep_interval = num_samples_per_sweep * 1/samp_rate
    if sweep_interval > chunk_size:
        raise ValueError((f"Sweep interval must be less than the chunk size. "
                          f"The computed sweep interval is {sweep_interval} [s], "
                          f"but the given chunk size is {chunk_size} [s]"))
    

def num_samples_per_step(samples_per_step: int, 
                         window_size: int) -> None:
    if window_size >= samples_per_step:
        raise ValueError((f"Window size must be strictly less than the number of samples per step. "
                          f"Got window size {window_size} [samples], which is more than or equal "
                          f"to the number of samples per step {samples_per_step}"))
    

def non_overlapping_steps(freq_step: float, 
                          samp_rate: int) -> None:
    if freq_step < samp_rate:
        raise NotImplementedError(f"SPECTRE does not yet support spectral steps overlapping in frequency. "
                                  f"Got frequency step {freq_step/1e6} [MHz] which is less than the sample "
                                  f"rate {samp_rate/1e6} [MHz]")
    

def step_interval(samples_per_step: int, 
                  samp_rate: int, 
                  api_latency: float) -> None:
    step_interval = samples_per_step * 1/samp_rate # [s]
    if step_interval < api_latency:
        warning_message = (f"The computed step interval of {step_interval} [s] is of the order of empirically "
                           f"derived api latency {api_latency} [s]; you may experience undefined behaviour!")
        warnings.warn(warning_message)
        _LOGGER.warning(warning_message)
    

def watch_extension(watch_extension: str,
                    target_extension: str) -> None:
    if watch_extension != target_extension:
        raise ValueError((f"Expected {target_extension} for the watch extension. "
                          f"Got {watch_extension}"))



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime

import numpy as np


def is_close(ar: np.ndarray, 
             ar_comparison: np.ndarray,
             absolute_tolerance: float) -> bool:
    """Close enough accounts for wiggle-room equating floats."""
    return np.all(np.isclose(ar, 
                             ar_comparison, 
                             atol=absolute_tolerance))


def find_closest_index(
    target_value: float | datetime, 
    array: np.ndarray, 
    enforce_strict_bounds: bool = False
) -> int:
    # Ensure input array is a numpy array
    array = np.asarray(array)

    # Convert to datetime64 if necessary
    if isinstance(target_value, datetime) or np.issubdtype(array.dtype, np.datetime64):
        target_value = np.datetime64(target_value)
        array = array.astype('datetime64[ns]')
    else:
        target_value = float(target_value)
        array = array.astype(float)

    # Check bounds if strict enforcement is required
    if enforce_strict_bounds:
        max_value, min_value = np.nanmax(array), np.nanmin(array)
        if target_value > max_value:
            raise ValueError(f"Target value {target_value} exceeds max array value {max_value}")
        if target_value < min_value:
            raise ValueError(f"Target value {target_value} is less than min array value {min_value}")

    # Find the index of the closest value
    return np.argmin(np.abs(array - target_value))


def normalise_peak_intensity(array: np.ndarray) -> np.ndarray:
    return array/np.nanmax(array)


def compute_resolution(array: np.ndarray) -> float:
    # Check that the array is one-dimensional
    if array.ndim != 1:
        raise ValueError("Input array must be one-dimensional")
    
    if len(array) < 2:
        raise ValueError("Input array must contain at least two elements")
    
    # Calculate differences between consecutive elements.
    resolutions = np.diff(array)

    return np.nanmedian(resolutions)


def compute_range(array: np.ndarray) -> float:
    # Check that the array is one-dimensional
    if array.ndim != 1:
        raise ValueError("Input array must be one-dimensional")
    
    if len(array) < 2:
        raise ValueError("Input array must contain at least two elements")
    return array[-1] - array[0]


def subtract_background(array: np.ndarray, start_index: int, end_index: int) -> np.ndarray:
    array -= np.nanmean(array[start_index:end_index+1])
    return array


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# SPDX-License-Identifier: GPL-3.0
#
# GNU Radio Python Flow Graph
# Title: Not titled yet
# GNU Radio version: 3.10.1.1

# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import signal
from argparse import ArgumentParser
from typing import Any

from gnuradio import blocks
from gnuradio import gr
from gnuradio.filter import firdes
from gnuradio.fft import window
from gnuradio.eng_arg import eng_float, intx
from gnuradio import eng_notation
from gnuradio import spectre

from spectre_core.cfg import CHUNKS_DIR_PATH
from spectre_core.file_handlers.configs import CaptureConfig

class tagged_staircase(gr.top_block):

    def __init__(self, 
                 capture_config: CaptureConfig):
        gr.top_block.__init__(self, "tagged-staircase", catch_exceptions=True)

        ##################################################
        # Unpack capture config
        ##################################################
        tag = capture_config['tag']
        step_increment = capture_config['step_increment']
        samp_rate = capture_config['samp_rate']
        min_samples_per_step = capture_config['min_samples_per_step']
        max_samples_per_step = capture_config['max_samples_per_step']
        freq_step = capture_config['freq_step']
        chunk_size = capture_config['chunk_size']
        is_sweeping = True

        ##################################################
        # Blocks
        ##################################################
        self.spectre_tagged_staircase_0 = spectre.tagged_staircase(min_samples_per_step, 
                                                                   max_samples_per_step, 
                                                                   freq_step,
                                                                   step_increment, 
                                                                   samp_rate)
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(CHUNKS_DIR_PATH,
                                                                     tag, 
                                                                     chunk_size, 
                                                                     samp_rate, 
                                                                     is_sweeping,
                                                                     'rx_freq',
                                                                     0
                                                                     )
        self.blocks_throttle_0 = blocks.throttle(gr.sizeof_gr_complex*1, samp_rate, True)


        ##################################################
        # Connections
        ##################################################
        self.connect((self.blocks_throttle_0, 0), (self.spectre_batched_file_sink_0, 0))
        self.connect((self.spectre_tagged_staircase_0, 0), (self.blocks_throttle_0, 0))




def main(capture_config: CaptureConfig, 
         top_block_cls=tagged_staircase, 
         options=None):
    tb = top_block_cls(capture_config)

    def sig_handler(sig=None, frame=None):
        tb.stop()
        tb.wait()

        sys.exit(0)

    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    tb.start()

    tb.wait()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# SPDX-License-Identifier: GPL-3.0
#
# GNU Radio Python Flow Graph
# Title: tuner_1_fixed
# GNU Radio version: 3.10.1.1

# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import signal
from argparse import ArgumentParser
from typing import Any

from gnuradio import gr
from gnuradio.filter import firdes
from gnuradio.fft import window
from gnuradio.eng_arg import eng_float, intx
from gnuradio import eng_notation
from gnuradio import sdrplay3
from gnuradio import spectre

from spectre_core.cfg import CHUNKS_DIR_PATH
from spectre_core.file_handlers.configs import CaptureConfig


class tuner_1_fixed(gr.top_block):

    def __init__(self, 
                 capture_config: CaptureConfig):
        gr.top_block.__init__(self, "tuner_1_fixed", catch_exceptions=True)

        ##################################################
        # Unpack capture config
        ##################################################
        samp_rate = capture_config['samp_rate']
        tag = capture_config['tag']
        chunk_size = capture_config['chunk_size']
        center_freq = capture_config['center_freq']
        bandwidth = capture_config['bandwidth']
        IF_gain = capture_config['IF_gain']
        RF_gain = capture_config['RF_gain']
        is_sweeping = False

        ##################################################
        # Blocks
        ##################################################
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(CHUNKS_DIR_PATH, 
                                                                     tag, 
                                                                     chunk_size, 
                                                                     samp_rate, 
                                                                     is_sweeping)
        self.sdrplay3_rspduo_0 = sdrplay3.rspduo(
            '',
            rspduo_mode="Single Tuner",
            antenna="Tuner 1 50 ohm",
            stream_args=sdrplay3.stream_args(
                output_type='fc32',
                channels_size=1
            ),
        )
        self.sdrplay3_rspduo_0.set_sample_rate(samp_rate)
        self.sdrplay3_rspduo_0.set_center_freq(center_freq)
        self.sdrplay3_rspduo_0.set_bandwidth(bandwidth)
        self.sdrplay3_rspduo_0.set_antenna("Tuner 1 50 ohm")
        self.sdrplay3_rspduo_0.set_gain_mode(False)
        self.sdrplay3_rspduo_0.set_gain(IF_gain, 'IF')
        self.sdrplay3_rspduo_0.set_gain(RF_gain, 'RF')
        self.sdrplay3_rspduo_0.set_freq_corr(0)
        self.sdrplay3_rspduo_0.set_dc_offset_mode(False)
        self.sdrplay3_rspduo_0.set_iq_balance_mode(False)
        self.sdrplay3_rspduo_0.set_agc_setpoint(-30)
        self.sdrplay3_rspduo_0.set_rf_notch_filter(False)
        self.sdrplay3_rspduo_0.set_dab_notch_filter(False)
        self.sdrplay3_rspduo_0.set_am_notch_filter(False)
        self.sdrplay3_rspduo_0.set_biasT(False)
        self.sdrplay3_rspduo_0.set_debug_mode(False)
        self.sdrplay3_rspduo_0.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo_0.set_show_gain_changes(False)


        ##################################################
        # Connections
        ##################################################
        self.connect((self.sdrplay3_rspduo_0, 0), (self.spectre_batched_file_sink_0, 0))





def main(capture_config: CaptureConfig, 
         top_block_cls=tuner_1_fixed, 
         options=None):

    tb = top_block_cls(capture_config)

    def sig_handler(sig=None, frame=None):
        tb.stop()
        tb.wait()

        sys.exit(0)

    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    tb.start()

    tb.wait()




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


# Global dictionaries to hold the mappings
panels = {}
    
def register_panel(panel_name: str): 
    def decorator(cls):
        panels[panel_name] = cls
        return cls
    return decorator


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

from queue import Queue
from typing import Optional
from abc import ABC, abstractmethod
from math import floor

from watchdog.events import (
    FileSystemEventHandler, 
    FileCreatedEvent, 
)

from spectre_core.chunks.factory import get_chunk_from_tag
from spectre_core.file_handlers.configs import CaptureConfig
from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.spectrograms.transform import join_spectrograms
from spectre_core.spectrograms.transform import (
    time_average, 
    frequency_average
)


class BaseEventHandler(ABC, FileSystemEventHandler):
    def __init__(self, 
                 tag: str):
        self._tag = tag

        self._Chunk = get_chunk_from_tag(tag)

        self._capture_config = CaptureConfig(tag)

        self._watch_extension = self._capture_config.get("watch_extension")
        if self._watch_extension is None:
            raise KeyError("The watch extension has not been specified in the capture config")

        # attribute to store the next file to be processed 
        # (specifically, the absolute file path of the file)
        self._queued_file: Optional[str] = None

        # spectrogram cache stores spectrograms in memory
        # such that they can be periodically written to files
        # according to the joining time.
        self._spectrogram: Optional[Spectrogram] = None
        


    @abstractmethod
    def process(self, 
                absolute_file_path: str) -> None:
        """Process the file stored at the input absolute file path.
        
        To be implemented by derived classes.
        """


    def on_created(self, 
                   event: FileCreatedEvent):
        """Process a newly created batch file, only once the next batch is created.
        
        Since we assume that the batches are non-overlapping in time, this guarantees
        we avoid post processing a file while it is being written to. Files are processed
        sequentially, in the order they are created.
        """

        # the 'src_path' attribute holds the absolute path of the newly created file
        absolute_file_path = event.src_path
        
        # only 'notice' a file if it ends with the appropriate extension
        # as defined in the capture config
        if absolute_file_path.endswith(self._watch_extension):
            _LOGGER.info(f"Noticed {absolute_file_path}")
            
            # If there exists a queued file, try and process it
            if self._queued_file is not None:
                try:
                    self.process(self._queued_file)
                except Exception:
                    _LOGGER.error(f"An error has occured while processing {self._queued_file}",
                                  exc_info=True)
                     # flush any internally stored spectrogram on error to avoid lost data
                    self._flush_spectrogram()
                    # re-raise the exception to the main thread
                    raise
            
            # Queue the current file for processing next
            _LOGGER.info(f"Queueing {absolute_file_path} for post processing")
            self._queued_file = absolute_file_path


    def _average_in_time(self, 
                         spectrogram: Spectrogram) -> Spectrogram:
        _LOGGER.info("Averaging spectrogram in time")
        requested_time_resolution = self._capture_config['time_resolution'] # [s]
        if requested_time_resolution is None:
            raise KeyError(f"Time resolution has not been specified in the capture config")
        average_over = floor(requested_time_resolution/spectrogram.time_resolution) if requested_time_resolution > spectrogram.time_resolution else 1
        return time_average(spectrogram, average_over)
    
    
    def _average_in_frequency(self, 
                              spectrogram: Spectrogram) -> Spectrogram:
        _LOGGER.info("Averaging spectrogram in frequency")
        frequency_resolution = self._capture_config['frequency_resolution'] # [Hz]
        if frequency_resolution is None:
            raise KeyError(f"Frequency resolution has not been specified in the capture config")
        average_over = floor(frequency_resolution/spectrogram.frequency_resolution) if frequency_resolution > spectrogram.frequency_resolution else 1
        return frequency_average(spectrogram, average_over)
    

    def _join_spectrogram(self, 
                          spectrogram: Spectrogram) -> None:
        _LOGGER.info("Joining spectrogram")
        if self._spectrogram is None:
            self._spectrogram = spectrogram
        else:
            self._spectrogram = join_spectrograms([self._spectrogram, spectrogram])

        if self._spectrogram.time_range >= self._capture_config['joining_time']:
            self._flush_spectrogram()
    

    def _flush_spectrogram(self) -> None:
        if self._spectrogram:
            _LOGGER.info(f"Flushing spectrogram to file with chunk start time {self._spectrogram.chunk_start_time}")
            self._spectrogram.save()
            _LOGGER.info("Flush successful, resetting spectrogram cache")
            self._spectrogram = None # reset the cache


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

import os

from spectre_core.chunks.base import BaseChunk
from spectre_core.post_processing.base import BaseEventHandler
from spectre_core.post_processing.event_handler_register import register_event_handler

@register_event_handler("sweep")
class EventHandler(BaseEventHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.previous_chunk: BaseChunk = None # cache for previous chunk
        

    def process(self, 
                absolute_file_path: str):
        _LOGGER.info(f"Processing: {absolute_file_path}")
        file_name = os.path.basename(absolute_file_path)
        base_file_name, _ = os.path.splitext(file_name)
        chunk_start_time, _ = base_file_name.split('_')
        chunk = self._Chunk(chunk_start_time, self._tag)

        _LOGGER.info("Creating spectrogram")
        spectrogram = chunk.build_spectrogram(previous_chunk = self.previous_chunk)

        spectrogram = self._average_in_time(spectrogram)
        spectrogram = self._average_in_frequency(spectrogram)
        self._join_spectrogram(spectrogram)

        # if the previous chunk has not yet been set, it means we are processing the first chunk
        # so we don't need to handle the previous chunk
        if self.previous_chunk is None:
            # instead, only set it for the next time this method is called
            self.previous_chunk = chunk
            
        # otherwise the previous chunk is defined (and by this point has already been processed)
        else:
            bin_chunk = self.previous_chunk.get_file('bin')
            _LOGGER.info(f"Deleting {bin_chunk.file_path}")
            bin_chunk.delete()

            hdr_chunk = self.previous_chunk.get_file('hdr')
            _LOGGER.info(f"Deleting {hdr_chunk.file_path}")
            hdr_chunk.delete()

            # and reassign the current chunk to be used as the previous chunk at the next call of this method
            self.previous_chunk = chunk



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from typing import List, Optional, Tuple
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes

from spectre_core.spectrograms.spectrogram import Spectrogram
from spectre_core.plotting.base import BasePanel, CutsPanel
from spectre_core.plotting.format import DEFAULT_FORMATS
from spectre_core.plotting.factory import get_panel
from spectre_core.plotting.library.spectrogram.panel import Panel as SpectrogramPanel

class PanelStack:
    def __init__(self, 
                 time_type: str = "seconds", 
                 figsize: Tuple[int, int] = (10, 10)):
        self._time_type = time_type
        self._figsize = figsize
        self._panels: List[BasePanel] = []
        self._superimposed_panels: List[BasePanel] = []
        self._fig: Optional[Figure] = None
        self._axs: Optional[np.ndarray[Axes]] = None


    @property
    def time_type(self) -> str:
        return self._time_type


    @property
    def panels(self) -> List[BasePanel]:
        return sorted(self._panels, key=lambda panel: panel.x_axis_type)


    @property
    def fig(self) -> Optional[Figure]:
        return self._fig


    @property
    def axs(self) -> Optional[np.ndarray[Axes]]:
        return np.atleast_1d(self._axs)


    @property
    def num_panels(self) -> int:
        return len(self._panels)


    def add_panel(self, 
                  panel_name: str, 
                  spectrogram: Spectrogram,
                  *args, 
                  identifier: Optional[str] = None,
                  **kwargs) -> None:
        panel = get_panel(panel_name, 
                          spectrogram, 
                          self._time_type, 
                          *args, 
                          **kwargs)
        if identifier: 
            panel.identifier = identifier
        self._panels.append(panel)


    def superimpose_panel(self, 
                          panel_name: str, 
                          spectrogram: Spectrogram, 
                          *args, 
                          identifier: Optional[str] = None,
                          **kwargs) -> None:
        panel = get_panel(panel_name, 
                          spectrogram, 
                          self._time_type, 
                          *args, 
                          **kwargs)
        if identifier:
            panel.identifier = identifier
        self._superimposed_panels.append(panel)


    def _init_plot_style(self) -> None:
        plt.style.use(DEFAULT_FORMATS.style)
        plt.rc('font', size=DEFAULT_FORMATS.small_size)
        plt.rc('axes', titlesize=DEFAULT_FORMATS.medium_size, 
                       labelsize=DEFAULT_FORMATS.medium_size)
        plt.rc('xtick', labelsize=DEFAULT_FORMATS.small_size)
        plt.rc('ytick', labelsize=DEFAULT_FORMATS.small_size)
        plt.rc('legend', fontsize=DEFAULT_FORMATS.small_size)
        plt.rc('figure', titlesize=DEFAULT_FORMATS.large_size)


    def _create_figure_and_axes(self) -> None:
        self._fig, self._axs = plt.subplots(self.num_panels, 1, figsize=self._figsize, layout="constrained")


    def _assign_axes(self) -> None:
        shared_axes = {}
        for i, panel in enumerate(self.panels):
            panel.ax = self.axs[i]
            panel.fig = self._fig
            if panel.x_axis_type in shared_axes:
                panel.ax.sharex(shared_axes[panel.x_axis_type])
            else:
                shared_axes[panel.x_axis_type] = panel.ax


    def _overlay_cuts(self, cuts_panel: CutsPanel) -> None:
        """Given a cuts panel, finds any corresponding spectrogram panels and adds the appropriate overlay"""
        for panel in self.panels:
            is_corresponding_panel = isinstance(panel, SpectrogramPanel) and (panel.tag == cuts_panel.tag)
            if is_corresponding_panel:
                panel.overlay_cuts(cuts_panel)


    def _overlay_superimposed_panels(self) -> None:
        for super_panel in self._superimposed_panels:
            for panel in self._panels:
                if panel.name == super_panel.name and (panel.identifier == super_panel.identifier): # find the matching panels via panel names
                    super_panel.ax, super_panel.fig = panel.ax, self._fig # and superimpose via axes sharing
                    super_panel.draw()
                    if isinstance(super_panel, CutsPanel):
                        self._overlay_cuts(super_panel)


    def show(self) -> None:
        self._init_plot_style()
        self._create_figure_and_axes()
        self._assign_axes()
        last_panel_per_axis = {panel.x_axis_type: panel for panel in self.panels}
        for panel in self.panels:
            panel.draw()
            panel.annotate_y_axis()
            if panel == last_panel_per_axis[panel.x_axis_type]:
                panel.annotate_x_axis()
            else:
                panel.hide_x_axis_labels()
            if isinstance(panel, CutsPanel):
                        self._overlay_cuts(panel)


        self._overlay_superimposed_panels()
        plt.show()
